/**
 * EasyBeans
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:$
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.injection.impl;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.rmi.PortableRemoteObject;

import org.ow2.easybeans.asm.Type;
import org.ow2.easybeans.injection.api.ArchiveInjectionException;
import org.ow2.easybeans.resolver.api.EZBJNDIBeanData;
import org.ow2.easybeans.resolver.api.EZBJNDIData;
import org.ow2.easybeans.resolver.api.EZBRemoteJNDIResolver;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.builder.webserviceref.WebServiceRefBuilder;
import org.ow2.util.ee.builder.webserviceref.factory.WebServiceRefObjectFactory;
import org.ow2.util.ee.deploy.api.deployable.CARDeployable;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperException;
import org.ow2.util.ee.metadata.car.api.ICarDeployableMetadata;
import org.ow2.util.ee.metadata.car.impl.CarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.common.api.ICommonClassMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonFieldMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonMethodMetadata;
import org.ow2.util.ee.metadata.common.api.interfaces.ISharedMetadata;
import org.ow2.util.ee.metadata.common.api.struct.IJAnnotationResource;
import org.ow2.util.ee.metadata.common.api.struct.IJEjbEJB;
import org.ow2.util.ee.metadata.common.api.struct.IJavaxPersistenceUnit;
import org.ow2.util.ee.metadata.common.api.struct.IJaxwsWebServiceRef;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.metadata.structures.JClass;
import org.ow2.util.scan.api.metadata.structures.JField;

/**
 * This class manages the injection.
 * @author Florent Benoit
 */
public class ArchiveInjection {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ArchiveInjection.class);

    /**
     * Defines java.lang.Object class.
     */
    public static final String JAVA_LANG_OBJECT = "java/lang/Object";

    /**
     * Archive that will be used to analyze the classes (may be null).
     */
    private IArchive archive = null;

    /**
     * Analyze of the archive has been done ?
     */
    private boolean analyzed = false;

    /**
     * Remote JNDI Resolver (if present).
     */
    private EZBRemoteJNDIResolver jndiResolver = null;

    /**
     * Collection of metadata that have been analyzed.
     */
    private Collection<? extends ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata>> metadataCollection = null;

    /**
     * Default constructor.
     * @param archive the given archive on which injection is based.
     */
    public ArchiveInjection(final IArchive archive) {
        this.archive = archive;
    }

    /**
     * Constructor for users that already have the metadatas.
     * @param carMetadata the metadatas to use.
     */
    public ArchiveInjection(final ICarDeployableMetadata carMetadata) {
        this.metadataCollection = carMetadata.getCarClassMetadataCollection();
        this.analyzed = true;
    }

    /**
     * Analyze the archive.
     * @throws ArchiveInjectionException if analyze fails
     */
    public void analyze(final ClassLoader classLoader) throws ArchiveInjectionException {

        // Get Deployable
        IDeployable<?> deployable = null;
        try {
            deployable = DeployableHelper.getDeployable(this.archive);
        } catch (DeployableHelperException e) {
            throw new ArchiveInjectionException("Unable to get a deployable on archive '" + this.archive + "'.", e);
        }

        // Client Archive ?
        if (CARDeployable.class.isInstance(deployable)) {
            // Create metadata
            ICarDeployableMetadata carDeployableMetadata = null;
            try {
                carDeployableMetadata = new CarDeployableMetadataFactory().createDeployableMetadata(CARDeployable.class
                        .cast(deployable), classLoader);
            } catch (DeployableMetadataException e) {
                throw new ArchiveInjectionException("Unable to get metadata on archive '" + this.archive + "'.", e);
            }

            // Use only common part
            this.metadataCollection = carDeployableMetadata.getCarClassMetadataCollection();
        } else {
            // TODO : adapt for other archives
            throw new UnsupportedOperationException("Can only manage injection for Client Deployable. Deployable found is '"
                    + deployable + "'.");
        }

        // Ok, now that metadata are here,

        this.analyzed = true;
    }

    /**
     * Initialize stuff in the given class.
     * @param clazz the class to initialize
     */
    public void init(final Class<?> clazz) throws ArchiveInjectionException {
        // Analyze the metadata if not done
        if (!this.analyzed) {
            analyze(clazz.getClassLoader());
        }

        // Get list of classes in the reverse order of the inheritance
        LinkedList<Class<?>> listClasses = new LinkedList<Class<?>>();
        Class<?> currentClass = clazz;
        while (currentClass != null) {
            listClasses.addFirst(currentClass);
            currentClass = currentClass.getSuperclass();
        }

        // Analyze classes and perform initialization
        for (Class<?> classToAnalyze : listClasses) {
            // Get the metadata of this class
            ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> classMetadata;
            classMetadata = getClassMetadata(classToAnalyze);

            // Found metadata ?
            if (classMetadata != null) {

                // Perform injection
                init(classMetadata, classToAnalyze, null);
            }
        }
    }

    /**
     * Perform injection in the given instance (or if null on a static way on
     * the class) with the provided metadata.
     * @param commonMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for
     *        static case.
     */
    public void init(final ICommonClassMetadata commonMetadata, final Class<?> clazz, final Object instance)
            throws ArchiveInjectionException {

        // Loop on the fields
        Collection<? extends ICommonFieldMetadata> fieldCollection = commonMetadata.getStandardFieldMetadataCollection();
        Iterator<? extends ICommonFieldMetadata> itField = fieldCollection.iterator();
        while (itField.hasNext()) {
            ICommonFieldMetadata fieldMetadata = itField.next();

            // Set the interface name by using the descriptor of the field
            Type typeInterface = Type.getType(fieldMetadata.getJField().getDescriptor());
            String interfaceName = typeInterface.getClassName();

            // Inject EJB
            IJEjbEJB jejb = fieldMetadata.getJEjbEJB();
            if (jejb != null) {
                // Update interface name
                jejb.setBeanInterface(interfaceName);
                // inject field
                injectFieldEJB(jejb, fieldMetadata, clazz, instance);
            }

            // Inject Resources
            IJAnnotationResource jAnnotationResource = fieldMetadata.getJAnnotationResource();
            if (jAnnotationResource != null) {
                // Update interface name
                jAnnotationResource.setType(interfaceName);

                // For some special case, fix the JNDI name to use
                if ("javax.transaction.UserTransaction".equals(jAnnotationResource.getType())) {
                    jAnnotationResource.setMappedName("java:comp/UserTransaction");
                } else if ("org.omg.CORBA.ORB".equals(jAnnotationResource.getType())) {
                    jAnnotationResource.setMappedName("java:comp/ORB");
                }

                // inject field
                injectFieldResource(jAnnotationResource, fieldMetadata, clazz, instance);
            }

            // Inject PersistenceUnit
            IJavaxPersistenceUnit jJavaxPersistenceUnit = fieldMetadata.getJavaxPersistenceUnit();
            if (jJavaxPersistenceUnit != null) {
                // inject field
                injectFieldPersistenceUnit(jJavaxPersistenceUnit, fieldMetadata, clazz, instance);
            }

            // Inject WebServiceRef
            IJaxwsWebServiceRef jaxwsWSR = fieldMetadata.getJaxwsWebServiceRef();
            if (jaxwsWSR != null) {

                // Update type name
                jaxwsWSR.setType(interfaceName);

                // Inject field
                injectWebServiceRef(jaxwsWSR, fieldMetadata, clazz, instance);
            }
        }

        // Visit metadatas at class-level
        List<IJaxwsWebServiceRef> webServiceRefs = commonMetadata.getJaxwsWebServiceRefs();
        if (webServiceRefs != null) {
            for (IJaxwsWebServiceRef ref : webServiceRefs) {

                // Inject class
                injectWebServiceRef(ref, commonMetadata, clazz, instance);

            }
        }

        // PostConstruct methods
        List<ICommonMethodMetadata> postConstructMethods = new ArrayList<ICommonMethodMetadata>();

        // Loop on the methods
        Collection<? extends ICommonMethodMetadata> methodCollection = commonMetadata.getStandardMethodMetadataCollection();
        Iterator<? extends ICommonMethodMetadata> itMethod = methodCollection.iterator();
        while (itMethod.hasNext()) {
            ICommonMethodMetadata methodMetadata = itMethod.next();

            // PostConstruct Method ?
            if (methodMetadata.isPostConstruct()) {
                postConstructMethods.add(methodMetadata);
            }

        }

        // Call Post Construct methods
        invokePostConstructMethods(postConstructMethods, clazz, instance);
    }

    /**
     * Perform injection in the given instance (or if null on a static way on
     * the class) with the provided metadata.
     * @param jaxwsWSR the service-ref that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for
     *        static case.
     */
    private void injectWebServiceRef(final IJaxwsWebServiceRef jaxwsWSR, final ISharedMetadata fieldMetadata,
            final Class clazz, final Object instance) throws ArchiveInjectionException {

        WebServiceRefBuilder builder = new WebServiceRefBuilder();
        Reference ref = null;
        try {
            ref = builder.build(jaxwsWSR);
        } catch (NamingException e) {
            throw new ArchiveInjectionException("Cannot build the Reference for " + "@WebServiceRef(" + jaxwsWSR.getName()
                    + ")", e);
        }

        String encName = jaxwsWSR.getName();

        // Now we have the instance ready to be injected
        // Inject only if metadata is a FieldMetadata
        // Because we only manage Field injection, not Method injection
        if (fieldMetadata instanceof ICommonFieldMetadata) {

            // Needs to set a TCCL because the ObjectFactory assumes that
            // it's used within the application's context
            ClassLoader loader = clazz.getClassLoader();
            ClassLoader old = Thread.currentThread().getContextClassLoader();
            try {
                Thread.currentThread().setContextClassLoader(loader);

                // Create the ObjectFactory and use it
                WebServiceRefObjectFactory factory = new WebServiceRefObjectFactory();
                Object injectedValue = factory.getObjectInstance(ref, null, null, null);

                // Inject the value
                setFieldValue((ICommonFieldMetadata) fieldMetadata, clazz, instance, injectedValue);

                // Update the JNDI binding name
                encName = buildENCName(jaxwsWSR.getName(), (ICommonFieldMetadata) fieldMetadata);
            } catch (Exception e) {
                throw new ArchiveInjectionException("Cannot inject " + "@WebServiceRef(" + jaxwsWSR.getName() + ")", e);
            } finally {
                Thread.currentThread().setContextClassLoader(old);
            }
        }

        // Bind the Reference in the JNDI
        bindValue(encName, ref);

    }

    /**
     * Call postconstruct methods.
     * @param postConstructMethods the methods to call
     * @param clazz the class to use for getting the methods
     * @param instance the given instance to call
     * @throws ArchiveInjectionException if invocation on the methods is not
     *         done
     */
    protected void invokePostConstructMethods(final List<ICommonMethodMetadata> postConstructMethods, final Class clazz,
            final Object instance) throws ArchiveInjectionException {
        // null list, return
        if (postConstructMethods == null) {
            return;
        }

        // Call each method
        for (ICommonMethodMetadata postConstructMethodMetadata : postConstructMethods) {
            Method postConstructMethod = null;
            try {
                postConstructMethod = clazz.getDeclaredMethod(postConstructMethodMetadata.getJMethod().getName());
            } catch (NoSuchMethodException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            }

            boolean accessible = postConstructMethod.isAccessible();
            try {
                postConstructMethod.setAccessible(true);
                postConstructMethod.invoke(instance);
            } catch (IllegalAccessException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            } catch (InvocationTargetException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            } finally {
                postConstructMethod.setAccessible(accessible);
            }
        }

    }

    /**
     * Perform injection in the given instance (or if null on a static way on
     * the class) with the provided metadata.
     * @param ejb the EJB-REF that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for
     *        static case.
     */
    public void injectFieldEJB(final IJEjbEJB ejb, final ICommonFieldMetadata fieldMetadata, final Class clazz,
            final Object instance) throws ArchiveInjectionException {
        // get data for the EJB
        String interfaceName = ejb.getBeanInterface();
        String beanName = ejb.getBeanName();
        String mappedName = ejb.getMappedName();

        // Check if there is a value bound in the ENC
        String encName = buildENCName(ejb.getName(), fieldMetadata);
        boolean needToBind = true;
        Object value = null;
        try {
            value = new InitialContext().lookup("java:comp/env/" + encName);
            needToBind = false;
        } catch (NamingException e) {
            logger.debug("Value for ENC '{0}' not bound in the registry", encName);
        }

        // Inject this EJB reference

        // JNDI Name to use if no value yet
        if (value == null) {
            String jndiName = null;

            // Mapped name defined ? use it as JNDI name
            if (mappedName != null) {
                jndiName = mappedName;
            } else {
                // needs to ask the remote side
                jndiName = getRemoteEJBJNDIName(interfaceName, beanName);
            }

            // OK, now that JNDI name is here, get value
            try {
                value = new InitialContext().lookup(jndiName);
            } catch (NamingException e) {
                throw new ArchiveInjectionException("Cannot get object with JNDI Name '" + jndiName + "' for ejb '" + ejb
                        + "'.", e);
            }
        }

        // Set value
        setFieldValue(fieldMetadata, clazz, instance, value);

        // bind in ENC
        if (needToBind) {
            bindValue(encName, value);
        }

    }

    /**
     * Perform injection in the given instance (or if null on a static way on
     * the class) with the provided metadata.
     * @param jJavaxPersistenceUnit the persistence-unit-ref that needs to be
     *        injecte
     * @param fieldMetadata the metadata that will be used for injection
     * @param clazz the class to use
     * @param instance the instance to use for the injection. May be null for
     *        static case.
     * @throws ArchiveInjectionException
     */
    public void injectFieldPersistenceUnit(final IJavaxPersistenceUnit jJavaxPersistenceUnit,
            final ICommonFieldMetadata fieldMetadata, final Class clazz, final Object instance)
            throws ArchiveInjectionException {
        String name = jJavaxPersistenceUnit.getName();
        Object value = null;
        // Check if there is a value bound in the ENC
        String encName = buildENCName(jJavaxPersistenceUnit.getName(), fieldMetadata);
        boolean needToBind = true;
        try {
            value = new InitialContext().lookup("java:comp/env/" + encName);
            needToBind = false;
        } catch (NamingException e) {
            logger.debug("Value for ENC '{0}' not bound in the registry", encName);
        }

        // No value, needs to find it
        if (value == null) {
            try {
                value = new InitialContext().lookup(name);
            } catch (NamingException e) {
                throw new ArchiveInjectionException("Cannot get object with JNDI Name '" + name + "' for Resource '"
                        + jJavaxPersistenceUnit + "'.", e);
            }
        }

        // Set value
        setFieldValue(fieldMetadata, clazz, instance, value);

        // bind in ENC
        if (needToBind) {
            bindValue(encName, value);
        }
    }

    /**
     * Perform injection in the given instance (or if null on a static way on
     * the class) with the provided metadata.
     * @param jAnnotationResource the resource-ref that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for
     *        static case.
     */
    public void injectFieldResource(final IJAnnotationResource jAnnotationResource, final ICommonFieldMetadata fieldMetadata,
            final Class clazz, final Object instance) throws ArchiveInjectionException {
        // get data for the Resource
        String mappedName = jAnnotationResource.getMappedName();

        // Value that will be injected
        Object value = null;

        // Check if there is a value bound in the ENC
        String encName = buildENCName(jAnnotationResource.getName(), fieldMetadata);
        boolean needToBind = true;
        try {
            value = new InitialContext().lookup("java:comp/env/" + encName);
            needToBind = false;
        } catch (NamingException e) {
            logger.debug("Value for ENC '{0}' not bound in the registry", encName);
        }

        // No value, needs to find it
        if (value == null) {
            if (mappedName == null) {
                String messageDestinationLink = jAnnotationResource.getMessageDestinationLink();
                if (messageDestinationLink == null) {
                    logger.warn("Injection of @Resource for {0} is not supported", jAnnotationResource);
                    return;
                }
                // Use MessageDestinationLink if presetn !
                mappedName = getRemoteMessageDestinationJNDIName(messageDestinationLink);
            }

            // OK, now that JNDI name is here, get value
            try {
                value = new InitialContext().lookup(mappedName);
            } catch (NamingException e) {
                throw new ArchiveInjectionException("Cannot get object with JNDI Name '" + mappedName + "' for Resource '"
                        + jAnnotationResource + "'.", e);
            }
        }

        // Set value
        setFieldValue(fieldMetadata, clazz, instance, value);

        // bind in ENC
        if (needToBind) {
            bindValue(encName, value);
        }

    }

    /**
     * Bind the given value in the ENC.
     * @param encName the ENC name
     * @param value the ENC value
     */
    protected void bindValue(final String encName, final Object value) {
        // bind in ENC if not already done
        try {
            new InitialContext().bind("java:comp/env/" + encName, value);
        } catch (NamingException e) {
            logger.warn("Unable to bind the value for ENC '{0}'.", encName);
        }
    }

    /**
     * Sets the value on a given field by using given arguments.
     * @param fieldMetadata the metadata for the name of the field
     * @param clazz the class of the field
     * @param instance the instance of the class
     * @param value the value to set
     * @throws ArchiveInjectionException if the value cannot be set
     */
    protected void setFieldValue(final ICommonFieldMetadata fieldMetadata, final Class clazz, final Object instance,
            final Object value) throws ArchiveInjectionException {

        // Value is obtained, set it
        // Get Field from the class
        JField jField = fieldMetadata.getJField();
        Field field = null;
        try {
            field = clazz.getDeclaredField(jField.getName());
        } catch (NoSuchFieldException e) {
            throw new ArchiveInjectionException("Cannot get field '" + jField + "' on the class '" + clazz.getName() + "'", e);
        }

        boolean accessible = field.isAccessible();
        try {
            // needs to be allowed to change value of the field
            field.setAccessible(true);

            // Set value
            try {
                field.set(instance, value);
            } catch (IllegalAccessException e) {
                throw new ArchiveInjectionException("Cannot set given value on the field '" + jField + "'.", e);
            }
        } finally {
            // set back accessible attribute
            field.setAccessible(accessible);
        }

    }

    /**
     * Build an ENC name for a given field.
     * @param name the existing name of the ENC specified through annotation/xml
     * @param fieldMetadata the metadata of the field
     * @return the updated enc name if there was none
     */
    protected String buildENCName(final String name, final ICommonFieldMetadata fieldMetadata) {
        String encName = name;
        if (encName == null || "".equals(encName)) {
            encName = fieldMetadata.getClassMetadata().getJClass().getName().replace("/", ".") + "/"
                    + fieldMetadata.getJField().getName();
        }
        return encName;
    }

    /**
     * Gets the JNDI resolver and initialize it if not already initialized.
     * @return JNDI Resolver
     * @throws ArchiveInjectionException if resolver is not available
     */
    protected EZBRemoteJNDIResolver getJNDIResolver() throws ArchiveInjectionException {
        if (this.jndiResolver == null) {
            // Get object
            Object o = null;
            try {
                o = new InitialContext().lookup("EZB_Remote_JNDIResolver");
            } catch (NamingException e) {
                // No Remote JNDI resolver, so throws exception
                throw new ArchiveInjectionException("No Remote EJB3 JNDI Resolver found", e);
            }

            // Object is here, so we can cast it
            this.jndiResolver = (EZBRemoteJNDIResolver) PortableRemoteObject.narrow(o, EZBRemoteJNDIResolver.class);
        }
        return this.jndiResolver;
    }

    /**
     * Gets a JNDI name from the server for a given message destination name.
     * @param messageDestinationName the name of the message destination
     * @return a JNDI name if value has been found, else an exception
     * @throws ArchiveInjectionException if JNDI Name is not found
     */
    protected String getRemoteMessageDestinationJNDIName(final String messageDestinationName) throws ArchiveInjectionException {

        String jndiName = null;

        // Ask the Resolver with interface name and bean name
        List<EZBJNDIData> jndiDataList = null;
        try {
            jndiDataList = getJNDIResolver().getMessageDestinationJNDINames(messageDestinationName);
        } catch (RemoteException re) {
            // No Remote JNDI resolver, so throw first exception
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + messageDestinationName + "'", re);
        }

        // Data is here, check if it is empty or not
        if (jndiDataList.size() == 0) {
            throw new ArchiveInjectionException("Unable to get JNDI Name for message destination '" + messageDestinationName
                    + "', no data was found on the remote side.");
        } else if (jndiDataList.size() > 1) {
            // too many entries
            logger.warn("There may be a problem for message destination '" + messageDestinationName + "', too many answers : '"
                    + jndiDataList + "'. Using the first entry");
        }

        // Only one item found, so get JNDI Name from this object
        EZBJNDIData jndiData = jndiDataList.get(0);
        // Update JNDI name
        jndiName = jndiData.getName();
        if (logger.isDebugEnabled()) {
            logger.debug("Found JNDI Name '" + jndiName + "' for message destination '" + messageDestinationName
                    + "', answers : '" + jndiDataList + "'.");
        }

        return jndiName;
    }

    /**
     * Gets a JNDI name from the server for a given bean/interface.
     * @param interfaceName the name of the interface
     * @param beanName the name of the bean
     * @return a JNDI name if value has been found, else an exception
     * @throws ArchiveInjectionException if JNDI Name is not found
     */
    protected String getRemoteEJBJNDIName(final String interfaceName, final String beanName) throws ArchiveInjectionException {

        String jndiName = null;

        // Ask the Resolver with interface name and bean name
        List<EZBJNDIBeanData> jndiDataList = null;
        try {
            jndiDataList = getJNDIResolver().getEJBJNDINames(interfaceName, beanName);
        } catch (RemoteException re) {
            // No Remote JNDI resolver, so throw first exception
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName + "'", re);
        }

        // Data is here, check if it is empty or not
        if (jndiDataList.size() == 0) {
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName
                    + "', no data was found on the remote side.");
        } else if (jndiDataList.size() > 1) {
            // too many entries
            logger.warn("There may be a problem for bean '" + interfaceName + "'/'" + beanName + "', too many answers : '"
                    + jndiDataList + "'. Using the first entry");
        }

        // Only one item found, so get JNDI Name from this object
        EZBJNDIBeanData jndiData = jndiDataList.get(0);
        // Update JNDI name
        jndiName = jndiData.getName();
        if (logger.isDebugEnabled()) {
            logger.debug("Found JNDI Name '" + jndiName + "' for '" + interfaceName + "'/'" + beanName
                    + "', answers : '" + jndiDataList + "'.");
        }

        return jndiName;
    }

    /**
     * Gets the metadata for a given class.
     * @param clazz the class on which we want to get metadata
     * @return the metadata if found else null
     */
    protected ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> getClassMetadata(
            final Class clazz) {
        return getClassMetadata(clazz.getName());
    }

    /**
     * Gets the metadata for a given class.
     * @param requestedClassName the name of the class on which we want to get
     *        metadata
     * @return the metadata if found else null
     */
    protected ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> getClassMetadata(
            final String requestedClassName) {

        // Get an iterator
        Iterator<? extends ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata>> it = this.metadataCollection
                .iterator();

        // Search the class by iterating on the metadata collection
        while (it.hasNext()) {
            // Declare temporary var
            ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> tmpMetadata;

            // Get current metadata
            tmpMetadata = it.next();

            // Metadata for the expected class ?
            JClass jClass = tmpMetadata.getJClass();
            if (jClass != null) {
                String className = jClass.getName();
                // compare name
                if (requestedClassName.equals(className.replace('/', '.'))) {
                    return tmpMetadata;
                }

            }
        }

        // Not found
        return null;
    }

}
